/*
 * network-manager-openvpn - OpenVPN integration with NetworkManager
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Copyright (C) 2008 Dan Williams, <dcbw@redhat.com>
 * Copyright (C) 2008 - 2018 Red Hat, Inc.
 */

#ifndef _IMPORT_EXPORT_H_
#define _IMPORT_EXPORT_H_

extern const char *_nmovpn_test_temp_path;

gboolean _nmovpn_test_args_parse_line (const char *line,
                                       gsize line_len,
                                       const char ***out_p,
                                       char **out_error);

NMConnection *do_import (const char *path, const char *contents, gsize contents_len, GError **error);

gboolean do_export (const char *path, NMConnection *connection, GError **error);

#endif
